unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls, ComCtrls, Mask, Menus, Grids, ShellAPI,
  llCSV, llBool, llMsgDlg, llFiles, llFileUtils, llDates, llMRU, llSettings,
  llStrings, FindUnit;

const
  AppVersion = '4.5';
  AppRegKey = 'Software\Ray Marron\Checkbook';
  WM_FIRSTTIME = WM_APP + 1;
  BlankSetting = '-';
  CheckFileExt = '.cb';
  MultiDelim = ';'; // Delimiter for multi-categories
  DepCat = '+++'; // Report categories
  UnkCat = 'UNK';
  NetCat = '===';
  ciDat = 0; // Column Index constants
  ciChk = 1;
  ciCat = 2;
  ciTxt = 3;
  ciRec = 4;
  ciPmt = 5;
  ciDep = 6;
  ciBal = 7;

{ Version history:

  4.5 2022-08-16
    * Made some costmetic fixes for High DPI displays and scaling.
  4.4 2021-04-23
    + Auto-save column widths on exit.
      (Set SaveColWidths=0 in registry to turn off)
    + Preferences: Added the "Reset column widths to factory default" button.
    + Preferences: Added the "Website" button.
  4.3 2020-07-16
    + Made the main window wider for first-time users so the date column won't
      have any issues appearing.
  4.2 2020-07-03
    + Added the Checkbook to Text report.
  4.1 2020-06-22
    + The Category Detail report would throw an error if you forgot to add the
      final remainder category with no equal sign on a multi-category entry.
      Categories are checked for errors when recalculating the balance and when
      running the detail report.
  4.0 2020-04-20
    + Added the ability to specify multiple categories for a single entry.
      Format: Cat1=amt1[;Cat2=amt2...];CatN (CatN has no amount == remainder)
      Example: Car=1860.00;Ent=211.50;Misc
      Handy for credit card payments that may include multiple things.
  3.1 2018-12-08
    + Added the "Treat deposits with a category as credits/refunds on reports"
      preference. When on, these amounts are subtracted from the category total.
      When off (old behavior), they are grouped with the rest of the deposits.
  3.0 2018-10-15
    + Recompiled in Delphi 10.2
  2.4 2017-02-12
    + Added a Welcome screen to help create a file on first-time use.
  2.32 2017-01-22
    + Added monthly averages to summmary report
  2.31 2016-04-27
    + Made the taskbar title "Checkbook" and the name of the current open file.
  2.3 2016-02-16
    + Changed original Category Report to Category Summary Report.
    + Added new Category Detail Report.
    + Added Ctrl-Z as the File|Reload shortcut key (Undo).
  2.2 2016-02-09
    + Improved Split error checking/messages.
    + Simplified default split date calc.
    + Added Initial Balance Entry preference, default "Beginning Balance".
  2.1 2016-02-06
    + Added File|Split to split the current checkbook file at the start of a
      new year (or any date, but it defaults to 12/31).
  2.0 2015-08-14
    + Started displaying the version in the title bar and keeping history.
  2015-08-13
    * Fixed incorrectly requiring a category on deposits.
  2015-07-27
    + Added File|Reload to main menu.
  2015-07-22
    + Added the category column and report.
  2013-2015
    + Numerous previous released and unreleased builds.
}

type
  TStringGridHack = class(TStringGrid)
  public
    // procedure MoveColumn(FromIndex, ToIndex: Longint);
    procedure MoveRow(FromIndex, ToIndex: Longint);
  end;

type
  TMainForm = class(TForm)
    AmountEdit: TEdit;
    BackBottomPanel: TPanel;
    BottomPanel: TPanel;
    btnEnter: TButton;
    CatEdit: TEdit;
    cbClearFields: TCheckBox;
    cbDeposit: TCheckBox;
    CheckNoEdit: TEdit;
    DateEdit: TDateTimePicker;
    DescEdit: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    MainGrid: TStringGrid;
    MainMenu: TMainMenu;
    mmData: TMenuItem;
    mmDataAutoSort: TMenuItem;
    mmDataFind: TMenuItem;
    mmDataFindNext: TMenuItem;
    mmDataRecalc: TMenuItem;
    mmDataReconMode: TMenuItem;
    mmDataSep0: TMenuItem;
    mmDataSortDate: TMenuItem;
    mmFile: TMenuItem;
    mmFileExit: TMenuItem;
    mmFileNew: TMenuItem;
    mmFileOpen: TMenuItem;
    mmFilePrefs: TMenuItem;
    mmFileRecent: TMenuItem;
    mmFileReload: TMenuItem;
    mmFileSep0: TMenuItem;
    mmFileSep1: TMenuItem;
    mmFileSplit: TMenuItem;
    mmReport: TMenuItem;
    mmReportCatDetail: TMenuItem;
    mmReportCatSummary: TMenuItem;
    mmReportText: TMenuItem;
    OpenDialog: TOpenDialog;
    pmgDelete: TMenuItem;
    pmgMoveDown: TMenuItem;
    pmgMoveUp: TMenuItem;
    pmgRecon: TMenuItem;
    pmMainGrid: TPopupMenu;
    SaveDialog: TSaveDialog;
    StatusBar: TStatusBar;
    procedure AmountEditExit(Sender: TObject);
    procedure btnEnterClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure MainGridDblClick(Sender: TObject);
    procedure MainGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure MainGridGetEditText(Sender: TObject; ACol, ARow: Integer;
      var Value: string);
    procedure MainGridKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure MainGridKeyPress(Sender: TObject; var Key: Char);
    procedure MainGridMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MainGridSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: string);
    procedure mmFileExitClick(Sender: TObject);
    procedure mmFileNewClick(Sender: TObject);
    procedure mmFileOpenClick(Sender: TObject);
    procedure mmFilePrefsClick(Sender: TObject);
    procedure mmFileReloadClick(Sender: TObject);
    procedure mmFileSplitClick(Sender: TObject);
    procedure mmDataFindNextClick(Sender: TObject);
    procedure mmDataFindClick(Sender: TObject);
    procedure mmDataAutoSortClick(Sender: TObject);
    procedure mmDataReconModeClick(Sender: TObject);
    procedure mmDataRecalcClick(Sender: TObject);
    procedure mmDataSortDateClick(Sender: TObject);
    procedure mmReportCatDetailClick(Sender: TObject);
    procedure mmReportCatSummaryClick(Sender: TObject);
    procedure mmReportTextClick(Sender: TObject);
    procedure pmMainGridPopup(Sender: TObject);
    procedure pmgDeleteClick(Sender: TObject);
    procedure pmgMoveDownClick(Sender: TObject);
    procedure pmgMoveUpClick(Sender: TObject);
    procedure pmgReconClick(Sender: TObject);
  private
    { Private declarations }
    FBackupDone: Boolean;
    FDirty: Boolean;
    FFindDate: TDateTime;
    FFindDateOpt: TFindDateOpt;
    FFindReconOpt: TFindReconOpt;
    FFindText: string;
    FMRU: TMRU;
    FSettings: TStringList;
    function GetAmount: Currency;
    function GetAutoSort: Boolean;
    function GetBackupCount: Integer;
    function GetBackupDate: TDateTime;
    function GetBackupDir: string;
    function GetCategoryRefunds: Boolean;
    function GetClearFields: Boolean;
    function GetColWidth(const ACol: Integer): Integer;
    function GetDefCat: string;
    function GetFindDir: TFindDir;
    function GetFindWrap: Boolean;
    function GetInitialEntry: string;
    function GetLastFile: string;
    function GetReconMode: Boolean;
    function GetRequireCat: Boolean;
    function GetTextDate1: TDateTime;
    function GetTextDate2: TDateTime;
    function GetTextStack: Boolean;
    function GetTextWidth: Integer;
    procedure SetAutoSort(const Value: Boolean);
    procedure SetBackupCount(const Value: Integer);
    procedure SetBackupDate(const Value: TDateTime);
    procedure SetBackupDir(const Value: string);
    procedure SetCategoryRefunds(const Value: Boolean);
    procedure SetClearFields(const Value: Boolean);
    procedure SetDefCat(const Value: string);
    procedure SetFindDir(const Value: TFindDir);
    procedure SetFindWrap(const Value: Boolean);
    procedure SetInitialEntry(const Value: string);
    procedure SetLastFile(const Value: string);
    procedure SetReconMode(const Value: Boolean);
    procedure SetRequireCat(const Value: Boolean);
    procedure SetTextDate1(const Value: TDateTime);
    procedure SetTextDate2(const Value: TDateTime);
    procedure SetTextStack(const Value: Boolean);
    procedure SetTextWidth(const Value: Integer);
    procedure AddEntry;
    procedure AutoMonthlyBackup(const AFile: string = '');
    function CellAmount(ACol, ARow: Integer): Currency;
    procedure CustomHint(Sender: TObject);
    procedure DoAutoSort;
    procedure DoFind;
    procedure DoFirstTime(var Msg: TMessage); message WM_FIRSTTIME;
    function FmtDate(const ADate: TDateTime): string;
    function FmtMoney(const AMoney: Currency): string;
    procedure LoadFile(const AFile: string = '');
    procedure MoveCurrentRowDown;
    procedure MoveCurrentRowUp;
    procedure RecalcAndSave;
    procedure ReportCategoryDetail(const AGroupByMonth, ASortByName: Boolean);
    procedure ReportCategorySummary;
    procedure ReportText;
    procedure ResetFields(const AResetDate: Boolean);
    procedure RollingBackup(const Force: Boolean = False);
    procedure SaveFile(const AFile: string = '');
    procedure SetColWidth(const ACol, AWidth: Integer);
    procedure SortDataByDate(const AForceRollingBackup: Boolean = False);
    procedure SplitFileByDate(const ADate: TDateTime; const AName: string);
  public
    { Public declarations }
    property AutoSort: Boolean read GetAutoSort write SetAutoSort;
    property BackupCount: Integer read GetBackupCount write SetBackupCount;
    property BackupDate: TDateTime read GetBackupDate write SetBackupDate;
    property BackupDir: string read GetBackupDir write SetBackupDir;
    property CategoryRefunds: Boolean read GetCategoryRefunds write SetCategoryRefunds;
    property ClearFields: Boolean read GetClearFields write SetClearFields;
    property DefaultCategory: string read GetDefCat write SetDefCat;
    property FindDir: TFindDir read GetFindDir write SetFindDir;
    property FindWrap: Boolean read GetFindWrap write SetFindWrap;
    property InitialEntry: string read GetInitialEntry write SetInitialEntry;
    property LastFile: string read GetLastFile write SetLastFile;
    property ReconMode: Boolean read GetReconMode write SetReconMode;
    property RequireCategory: Boolean read GetRequireCat write SetRequireCat;
    property TextDate1: TDateTime read GetTextDate1 write SetTextDate1;
    property TextDate2: TDateTime read GetTextDate2 write SetTextDate2;
    property TextStack: Boolean read GetTextStack write SetTextStack;
    property TextWidth: Integer read GetTextWidth write SetTextWidth;
    procedure AssignDefaultColWidths;
    procedure MRUClicked(Sender: TObject);
    procedure StoreCurrentColWidths;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

uses PrefsUnit, SplitUnit, TextUnit, Welcome;

{ TStringGridHack }

procedure TStringGridHack.MoveRow(FromIndex, ToIndex: Integer);
begin
  inherited;
end;

{ TMainForm }

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if FDirty then
    RecalcAndSave;
  ClearFields := cbClearFields.Checked;
  if StoB(FSettings.Values['SaveColWidths'], True) then
    StoreCurrentColWidths;
  SaveFormPos(AppRegKey, Self, FSettings);
  FSettings.Free;
  FMRU.Save;
  CanClose := True;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  Application.OnHint := CustomHint;
  FSettings := TStringList.Create;
  FMRU := TMRU.Create(mmFileRecent, AppRegKey);
  FMRU.OnClick := MRUClicked;
  FBackupDone := False;
  FDirty := False;
  FFindDate := Date;
  FFindDateOpt := fdAny;
  FFindReconOpt := frAny;
  // Default Settings (must be present before calling LoadWindowPos):
  AutoSort := True; // Default to automatically sort by date
  BackupCount := 10;
  BackupDate := NextMonthDay(Date, 1);
  BackupDir := '.';
  CategoryRefunds := True;
  ClearFields := True;
  DefaultCategory := BlankSetting;
  FindDir := fdUp;
  FindWrap := True;
  InitialEntry := 'Beginning Balance';
  LastFile := BlankSetting; // AddSlash(GetCurrentDir) + DtoS(Date, dtpYear) + CheckFileExt;
  RequireCategory := False;
  TextDate1 := 0;
  TextDate2 := 0;
  TextStack := True;
  TextWidth := 0;
  AssignDefaultColWidths;
  FSettings.Values['SaveColWidths'] := '1';
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  MainForm.Caption := MainForm.Caption + ' ' + AppVersion;
  LoadFormPos(AppRegKey, Self, FSettings);
  FMRU.RemoveDeletedFiles;
  mmDataAutoSort.Checked := AutoSort;
  StatusBar.Font := MainForm.Font;
  ResetFields(True);
  if LastFile = BlankSetting then begin
    LastFile := '';
    PostMessage(Handle, WM_FIRSTTIME, 0, 0);
  end else if LastFile <> '' then
    LoadFile;
end;

{ Property Getter/Setters }

function TMainForm.GetAutoSort: Boolean;
begin
  Result := StoB(FSettings.Values['AutoSort']);
end;

procedure TMainForm.SetAutoSort(const Value: Boolean);
begin
  FSettings.Values['AutoSort'] := BtoS(Value);
  mmDataAutoSort.Checked := Value;
end;

function TMainForm.GetBackupCount: Integer;
begin
  Result := StrToIntDef(FSettings.Values['BackupCount'], 10);
end;

procedure TMainForm.SetBackupCount(const Value: Integer);
begin
  FSettings.Values['BackupCount'] := IntToStr(Value);
end;

function TMainForm.GetBackupDate: TDateTime;
begin
  Result := StoDDef(FSettings.Values['AutoBackupDate'], 0);
end;

procedure TMainForm.SetBackupDate(const Value: TDateTime);
begin
  FSettings.Values['AutoBackupDate'] := DtoS(Value);
end;

function TMainForm.GetBackupDir: string;
begin
  Result := FSettings.Values['AutoBackupDir'];
end;

procedure TMainForm.SetBackupDir(const Value: string);
begin
  FSettings.Values['AutoBackupDir'] := Value;
end;

function TMainForm.GetCategoryRefunds: Boolean;
begin
  Result := StoB(FSettings.Values['CategoryRefunds'], True);
end;

procedure TMainForm.SetCategoryRefunds(const Value: Boolean);
begin
  FSettings.Values['CategoryRefunds'] := BtoS(Value);
end;

function TMainForm.GetClearFields: Boolean;
begin
  Result := StoB(FSettings.Values['ClearFields']);
end;

function TMainForm.GetColWidth(const ACol: Integer): Integer;
begin
  Result := StrToIntDef(FSettings.Values['ColWidth' + IntToStr(ACol)], 60);
end;

procedure TMainForm.SetClearFields(const Value: Boolean);
begin
  FSettings.Values['ClearFields'] := BtoS(Value);
end;

procedure TMainForm.SetColWidth(const ACol, AWidth: Integer);
begin
  FSettings.Values['ColWidth' + IntToStr(ACol)] := IntToStr(AWidth);
end;

function TMainForm.GetDefCat: string;
begin
  Result := FSettings.Values['DefaultCategory'];
  if Result = BlankSetting then
    Result := '';
end;

procedure TMainForm.SetDefCat(const Value: string);
begin
  // Need a non-blank value to keep the name/value pair in FSettings
  if IsBlank(Value) then
    FSettings.Values['DefaultCategory'] := BlankSetting
  else
    FSettings.Values['DefaultCategory'] := Value;
end;

function TMainForm.GetFindDir: TFindDir;
begin
  Result := TFindDir(StrToIntDef(FSettings.Values['FindDirection'], Ord(fdUp)));
end;

procedure TMainForm.SetFindDir(const Value: TFindDir);
begin
  FSettings.Values['FindDirection'] := IntToStr(Ord(Value));
end;

function TMainForm.GetFindWrap: Boolean;
begin
  Result := StoB(FSettings.Values['FindWrap']);
end;

procedure TMainForm.SetFindWrap(const Value: Boolean);
begin
  FSettings.Values['FindWrap'] := BtoS(Value);
end;

function TMainForm.GetInitialEntry: string;
begin
  Result := FSettings.Values['InitialEntry'];
end;

procedure TMainForm.SetInitialEntry(const Value: string);
begin
  if Value <> '' then
    FSettings.Values['InitialEntry'] := Value;
end;

function TMainForm.GetLastFile: string;
begin
  Result := FSettings.Values['LastFile'];
end;

procedure TMainForm.SetLastFile(const Value: string);
begin
  FSettings.Values['LastFile'] := Value;
end;

function TMainForm.GetReconMode: Boolean;
begin
  Result := mmDataReconMode.Checked; // Not tied to an FSettings value
end;

procedure TMainForm.SetReconMode(const Value: Boolean);
var
  bMode: Boolean;
  i: Integer;
begin
  if mmDataReconMode.Checked = Value then
    Exit; // No change!
  mmDataReconMode.Checked := Value;
  bMode := mmDataReconMode.Checked;
  if bMode then
    MainGrid.Options := MainGrid.Options + [goRowSelect] - [goEditing]
  else
    MainGrid.Options := MainGrid.Options - [goRowSelect] + [goEditing];
  // BottomPanel.Enabled := not bMode;  (Looks nicer when the controls grey out)
  for i := 0 to BottomPanel.ControlCount - 1 do
    BottomPanel.Controls[i].Enabled := not bMode;
  if bMode then
    MainGrid.SetFocus
  else
    DescEdit.SetFocus;
  CustomHint(nil);
end;

function TMainForm.GetRequireCat: Boolean;
begin
  Result := StoB(FSettings.Values['RequireCategory']);
end;

procedure TMainForm.SetRequireCat(const Value: Boolean);
begin
  FSettings.Values['RequireCategory'] := BtoS(Value);
end;

function TMainForm.GetTextDate1: TDateTime;
begin
  Result := StoDDef(DigitsOnly(FSettings.Values['TextDate1']), 0);
end;

procedure TMainForm.SetTextDate1(const Value: TDateTime);
begin
  if Value > 0 then
    FSettings.Values['TextDate1'] := FmtDate(Value)
  else
    FSettings.Values['TextDate1'] := BlankSetting;
end;

function TMainForm.GetTextDate2: TDateTime;
begin
  Result := StoDDef(DigitsOnly(FSettings.Values['TextDate2']), 0);
end;

procedure TMainForm.SetTextDate2(const Value: TDateTime);
begin
  if Value > 0 then
    FSettings.Values['TextDate2'] := FmtDate(Value)
  else
    FSettings.Values['TextDate2'] := BlankSetting;
end;

function TMainForm.GetTextStack: Boolean;
begin
  Result := StoB(FSettings.Values['TextStack']);
end;

procedure TMainForm.SetTextStack(const Value: Boolean);
begin
  FSettings.Values['TextStack'] := BtoS(Value);
end;

function TMainForm.GetTextWidth: Integer;
begin
  {
    >0 = Character count
     0 = Auto-fit all data
  }
  Result := StrToIntDef(FSettings.Values['TextWidth'], 120);
end;

procedure TMainForm.SetTextWidth(const Value: Integer);
begin
  FSettings.Values['TextWidth'] := IntToStr(Value);
end;

{ Control Events }

procedure TMainForm.AmountEditExit(Sender: TObject);
var
  a: Currency;
begin
  if TryStrToCurr(AmountEdit.Text, a) then
    AmountEdit.Text := FormatFloat('0.00', a)
  else
    AmountEdit.Text := '0.00';
end;

procedure TMainForm.AssignDefaultColWidths;
var
  i: Integer;
begin
  // Default column widths:
  SetColWidth(ciDat, 70);
  SetColWidth(ciChk, 60);
  SetColWidth(ciCat, 45);
  SetColWidth(ciTxt, 300);
  SetColWidth(ciRec, 18);
  SetColWidth(ciPmt, 60);
  SetColWidth(ciDep, 60);
  SetColWidth(ciBal, 60);
  for i := 0 to MainGrid.ColCount - 1 do
    MainGrid.ColWidths[i] := GetColWidth(i);
end;

procedure TMainForm.btnEnterClick(Sender: TObject);
begin
  AddEntry;
end;

procedure TMainForm.MainGridDblClick(Sender: TObject);
begin
  if ReconMode then
    pmgReconClick(Sender);
end;

procedure TMainForm.MainGridDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  s: string;
  SavedAlign: Word;
begin
  if ARow > 0 then begin
    // Right-justify the numeric entries
    if (ACol = ciPmt) or (ACol = ciDep) or (ACol = ciBal) then begin
      s := MainGrid.Cells[ACol, ARow];
      SavedAlign := SetTextAlign(MainGrid.Canvas.Handle, TA_RIGHT);
      // MainGrid.Canvas.FillRect(Rect);
      MainGrid.Canvas.TextRect(Rect, Rect.Right - 4, Rect.Top + 1, s);
      SetTextAlign(MainGrid.Canvas.Handle, SavedAlign);
    end;
  end;
end;

procedure TMainForm.MainGridGetEditText(Sender: TObject; ACol, ARow: Integer;
  var Value: string);
begin
  Value := MainGrid.Cells[ACol, ARow];
  if (ARow < 1) or (ACol = ciBal) then
    MainGrid.EditorMode := False; // Can't edit headers or balance
end;

procedure TMainForm.MainGridKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_F8 then
    pmgReconClick(Sender); // F8 toggles reconciled
end;

procedure TMainForm.MainGridKeyPress(Sender: TObject; var Key: Char);
begin
  if ReconMode and (Key = ' ') then begin
    pmgReconClick(Sender);
    if (MainGrid.Row + 1) < MainGrid.RowCount then
      MainGrid.Row := MainGrid.Row + 1;
  end;
end;

procedure TMainForm.MainGridMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  // Make right clicks select the row under the cursor:
  if Button = mbRight then
    MainGrid.Perform(WM_LBUTTONDOWN, 0, MakeLParam(Word(X), Word(Y)));
end;

procedure TMainForm.MainGridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
begin
  FDirty := True;
  { Can't tell when a user is finished editing a cell without dirty hacks. :(
    If I could, I would automatically sort or recalc when done.
  }
end;

procedure TMainForm.mmDataAutoSortClick(Sender: TObject);
begin
  AutoSort := not AutoSort;
end;

procedure TMainForm.mmDataFindClick(Sender: TObject);
var
  FF: TFindForm;
begin
  FF := TFindForm.Create(Self);
  try
    FF.FindText := FFindText;
    FF.FindDate := FFindDate;
    FF.FindDateOpt := FFindDateOpt;
    FF.FindReconOpt := FFindReconOpt;
    FF.FindDirection := FindDir;
    FF.FindWrap := FindWrap;
    if FF.ShowModal = mrOk then begin
      FFindText := FF.FindText;
      FFindDate := FF.FindDate;
      FFindDateOpt := FF.FindDateOpt;
      FFindReconOpt := FF.FindReconOpt;
      FindDir := FF.FindDirection;
      FindWrap := FF.FindWrap;
      DoFind;
    end;
  finally
    FF.Free;
  end;
end;

procedure TMainForm.mmDataFindNextClick(Sender: TObject);
begin
  if FFindText <> '' then
    DoFind
  else
    mmDataFindClick(Sender);
end;

procedure TMainForm.mmDataRecalcClick(Sender: TObject);
begin
  RecalcAndSave;
end;

procedure TMainForm.mmDataReconModeClick(Sender: TObject);
begin
  ReconMode := not ReconMode;
end;

procedure TMainForm.mmDataSortDateClick(Sender: TObject);
begin
  SortDataByDate;
end;

procedure TMainForm.mmFileExitClick(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.mmFileNewClick(Sender: TObject);
var
  sFile: string;
begin
  SaveDialog.InitialDir := ExtractFileDir(LastFile);
  SaveDialog.FileName :=
    ExtractFilePath(LastFile) + DtoS(Date, dtpYear) + CheckFileExt;
  if SaveDialog.Execute then begin
    sFile := SaveDialog.FileName;
    CreateEmptyFile(sFile);
    LoadFile(sFile);
  end;
end;

procedure TMainForm.mmFileOpenClick(Sender: TObject);
begin
  OpenDialog.InitialDir := ExtractFileDir(LastFile);
  if OpenDialog.Execute then begin
    if FDirty then
      RecalcAndSave;
    LoadFile(OpenDialog.FileName);
  end;
end;

procedure TMainForm.mmFilePrefsClick(Sender: TObject);
var
  PF: TPrefsForm;
begin
  PF := TPrefsForm.Create(Self);
  try
    PF.Settings := FSettings;
    PF.ShowModal;
  finally
    PF.Free;
  end;
end;

procedure TMainForm.mmFileReloadClick(Sender: TObject);
begin
  if LastFile = '' then
    Exit;
  if MsgYesNo('Reload the current file from the last save?',
              'Reload File', Self) then
    LoadFile;
end;

procedure TMainForm.mmFileSplitClick(Sender: TObject);
var
  F: TSplitForm;
begin
  F := TSplitForm.Create(Self);
  try
    if F.ShowModal = mrOk then
      SplitFileByDate(F.DateEdit.Date, F.NameEdit.Text);
  finally
    F.Free;
  end;
end;

procedure TMainForm.mmReportCatDetailClick(Sender: TObject);
var
  i: Integer;
  bMonthly, bSorted: Boolean;
begin
  i := MsgYesNoCancel('Group expenses by month?', 'Category Detail Report', Self);
  if i = mrCancel then
    Exit;
  bMonthly := (i = mrYes);
  i := MsgYesNoCancel('Sort entries by name?', 'Category Detail Report', Self);
  if i = mrCancel then
    Exit;
  bSorted := (i = mrYes);
  ReportCategoryDetail(bMonthly, bSorted);
end;

procedure TMainForm.mmReportCatSummaryClick(Sender: TObject);
begin
  ReportCategorySummary;
end;

procedure TMainForm.mmReportTextClick(Sender: TObject);
var
  F: TTextReportForm;
  y, m, d: Word;
  dtTemp: TDateTime;
begin
  DecodeDate(Now, y, m, d);
  F := TTextReportForm.Create(Self);
  try
    dtTemp := TextDate1;
    if dtTemp > 0 then begin
      F.StartDate.Date := dtTemp;
      F.StartDate.Checked := True;
    end else begin
      F.StartDate.Date := EncodeDate(y, 1, 1);
      F.StartDate.Checked := False;
    end;
    dtTemp := TextDate2;
    if dtTemp > 0 then begin
      F.EndDate.Date := dtTemp;
      F.EndDate.Checked := True;
    end else begin
      F.EndDate.Date := EncodeDate(y, 12, 31);
      F.EndDate.Checked := False;
    end;
    if TextWidth > 0 then begin
      F.WidthEdit.Value := TextWidth;
      F.cbAuto.Checked := False;
    end else
      F.cbAuto.Checked := True;
    F.cbStack.Checked := TextStack;

    if F.ShowModal = mrCancel then
      Exit;

    if F.StartDate.Checked then
      TextDate1 := F.StartDate.Date
    else
      TextDate1 := 0;
    if F.EndDate.Checked then
      TextDate2 := F.EndDate.Date
    else
      TextDate2 := 0;
    if F.cbAuto.Checked then
      TextWidth := 0
    else
      TextWidth := F.WidthEdit.Value;
    TextStack := F.cbStack.Checked;

    ReportText;

  finally
    F.Free;
  end;
end;

procedure TMainForm.pmgDeleteClick(Sender: TObject);
var
  i, r: Integer;
begin
  // Delete current entry
  if MainGrid.Row < 1 then
    Exit;
  if not MsgYesNo('Are you sure you want to delete the "' +
                  MainGrid.Cells[ciTxt, MainGrid.Row] + '" entry?',
                  'Delete Entry', Self) then
    Exit;
  // There is no delete row method, so we must move the following rows up and
  // then remove the last one:
  r := MainGrid.Row;
  for i := r to MainGrid.RowCount - 2 do
    MainGrid.Rows[i].Assign(MainGrid.Rows[i+1]);
  MainGrid.RowCount := MainGrid.RowCount - 1;
  RecalcAndSave;
end;

procedure TMainForm.pmgMoveDownClick(Sender: TObject);
begin
  MoveCurrentRowDown;
  RecalcAndSave;
end;

procedure TMainForm.pmgMoveUpClick(Sender: TObject);
begin
  MoveCurrentRowUp;
  RecalcAndSave;
end;

procedure TMainForm.pmgReconClick(Sender: TObject);
begin
  if MainGrid.Row < 1 then
    Exit;
  if MainGrid.Cells[ciRec, MainGrid.Row] = '*' then
    MainGrid.Cells[ciRec, MainGrid.Row] := ' '
  else
    MainGrid.Cells[ciRec, MainGrid.Row] := '*';
  FDirty := True;
end;

procedure TMainForm.pmMainGridPopup(Sender: TObject);
begin
  pmgRecon.Checked := MainGrid.Cells[ciRec, MainGrid.Row] = '*';
end;

{ Non-Event/Property Procedures }

procedure TMainForm.AddEntry;
var
  r: Integer;
  Bal: Currency;
  sDesc, sCat: string;
begin
  sDesc := Trim(DescEdit.Text);
  if (sDesc <> '') and (GetAmount > 0.00) then begin
    sCat := Trim(CatEdit.Text);
    if RequireCategory and (sCat = '') and not cbDeposit.Checked then begin
      CatEdit.SetFocus;
      Exit;
    end;
    if (MainGrid.RowCount > 2) or (MainGrid.Cells[ciDat, 1] <> '') then
      MainGrid.RowCount := MainGrid.RowCount + 1;
    r := MainGrid.RowCount - 1;
    MainGrid.Cells[ciDat, r] := FmtDate(DateEdit.DateTime);
    MainGrid.Cells[ciChk, r] := Trim(CheckNoEdit.Text);
    MainGrid.Cells[ciCat, r] := sCat;
    MainGrid.Cells[ciTxt, r] := sDesc;
    MainGrid.Cells[ciRec, r] := '';
    if MainGrid.RowCount = 2 then
      Bal := 0.00
    else
      Bal := CellAmount(ciBal, MainGrid.RowCount - 2);
    if cbDeposit.Checked then begin
      MainGrid.Cells[ciDep, r] := FmtMoney(GetAmount);
      Bal := Bal + GetAmount;
    end else begin
      MainGrid.Cells[ciPmt, r] := FmtMoney(GetAmount);
      Bal := Bal - GetAmount;
    end;
    MainGrid.Cells[ciBal, r] := FmtMoney(Bal);
    MainGrid.Row := r;
    MainGrid.Col := ciBal;
    if AutoSort then
      DoAutoSort;
    SaveFile;
    if cbClearFields.Checked then
      ResetFields(False);
    DescEdit.SetFocus;
  end;
end;

procedure TMainForm.AutoMonthlyBackup(const AFile: string);
var
  sPath, sFile, sBack: string;
  y, m, d: Word;
begin
  // Check preferences:
  sPath := BackupDir;
  if sPath = '-' then
    Exit; // Holdover from earlier version - don't do it when set to a dash
  if BackupDate < 1 then
    Exit; // 1899-12-30 means don't do one
  if Date < BackupDate then
    Exit; // Not time yet!
  // Make backup:
  if AFile = '' then
    sFile := LastFile
  else
    sFile := AFile;
  if Length(sPath) < 2 then
    sPath := ExtractFilePath(sFile); // If blank or . use the file's path
  // d:\path\file.20131101.cb
  sBack := AddSlash(sPath) + ExtractFileNameNoExt(sFile) +
           '.' + DtoS(Date) + ExtractFileExt(sFile);
  CopyFile(PChar(sFile), PChar(sBack), False);
  // Set the next backup date to the same day next month:
  // (A day > 28 will eventually be reduced to 28)
  DecodeDate(BackupDate, y, m, d);
  BackupDate := NextMonthDay(Date + 1, d);
end;

function TMainForm.CellAmount(ACol, ARow: Integer): Currency;
begin
  Result := StrToCurrDef(StringReplace(MainGrid.Cells[ACol, ARow], ',', '', [rfReplaceAll]), -1.00);
end;

procedure TMainForm.CustomHint(Sender: TObject);
begin
  if Application.Hint <> '' then
    StatusBar.SimpleText := Application.Hint
  else if ReconMode then
    StatusBar.SimpleText := LastFile + ' - Reconciliation Mode'
  else
    StatusBar.SimpleText := LastFile;
end;

procedure TMainForm.DoAutoSort;
var
  r: Integer;
  bRecalc: Boolean;
  Bal: Currency;
begin
  // The new entry is always the last row in the grid.
  // Move the entry up until the date above it is not greater than it.
  bRecalc := False;
  r := MainGrid.RowCount - 1;
  while (r > 1) and (MainGrid.Cells[ciDat, r - 1] > MainGrid.Cells[ciDat, r]) do begin
    MoveCurrentRowUp;
    r := r - 1;
    bRecalc := True;
  end;
  if bRecalc then begin
    // Just do a selective recalc from our new position to EOF:
    if r > 1 then
      Bal := CellAmount(ciBal, r - 1)
    else
      Bal := 0;
    while r < MainGrid.RowCount do begin
      if Length(MainGrid.Cells[ciPmt, r]) >= 4 {0.00} then
        Bal := Bal - CellAmount(ciPmt, r)
      else
        Bal := Bal + CellAmount(ciDep, r);
      MainGrid.Cells[ciBal, r] := FmtMoney(Bal);
      Inc(r);
    end;
  end;
end;

procedure TMainForm.DoFind;
var
  r, i: Integer;
  bFound: Boolean;
  sDate, sUpper: string;
begin
  if FFindText = '' then
    Exit;
  bFound := False;
  r := MainGrid.Row;
  sDate := FmtDate(FFindDate);
  sUpper := UpperCase(FFindText);
  repeat
    if FindDir = fdUp then begin
      if r > 1 then
        Dec(r)
      else if FindWrap then
        r := MainGrid.RowCount - 1
      else
        break;
    end else begin
      if r < MainGrid.RowCount - 1 then
        Inc(r)
      else if FindWrap then
        r := 1
      else
        break;
    end;
    if FFindReconOpt <> frAny then begin
      if (FFindReconOpt = frRec) and (MainGrid.Cells[ciRec, r] <> '*') then
        Continue;
      if (FFindReconOpt = frUnrec) and (MainGrid.Cells[ciRec, r] = '*') then
        Continue;
    end;
    if FFindDateOpt <> fdAny then begin
      if (FFindDateOpt = fdLess) and (MainGrid.Cells[ciDat, r] >= sDate) then
        Continue;
      if (FFindDateOpt = fdGreater) and (MainGrid.Cells[ciDat, r] <= sDate) then
        Continue;
    end;
    bFound := Pos(sUpper, UpperCase(MainGrid.Rows[r].CommaText)) > 0;
  until bFound or (r = MainGrid.Row);
  if bFound then begin
    MainGrid.Row := r;
    // Find and highlight the column where the value was found.
    for i := 0 to MainGrid.ColCount - 1 do begin
      if Pos(sUpper, UpperCase(MainGrid.Cells[i, r])) > 0 then begin
        MainGrid.Col := i;
        break;
      end;
    end;
    MainGrid.SetFocus;
  end else begin
    sUpper := '';
    if not FindWrap then begin
      if FindDir = fdUp then
        sUpper := 'Reached top - '
      else
        sUpper := 'Reached bottom - '
    end;
    sUpper := sUpper + 'No matching entries found.';
    MsgInfo(sUpper, 'Find', Self);
  end;
end;

procedure TMainForm.DoFirstTime(var Msg: TMessage);
var
  WF: TWelcomeForm;
  dtStart: TDateTime;
  Bal: Currency;
  slFile: TStringList;
begin
  WF := TWelcomeForm.Create(Self);
  try
    // Start with the last day of last month
    dtStart := Trunc(LastOfMonth(DateMath(Date, 0, -1, 0, 0, 0, 0, 0)));
    WF.FileEdit.Text :=
      AddSlash(GetCurrentDir) + DtoS(dtStart, dtpYear) + CheckFileExt;
    WF.DateEdit.DateTime := dtStart;
    WF.BalanceEdit.Text := '0.00';
    WF.DescEdit.Text := InitialEntry;
    if WF.ShowModal = mrOK then begin
      LastFile := WF.FileEdit.Text;
      Bal := StrToCurrDef(WF.BalanceEdit.Text, 0.00);
      if Bal = 0 then
        CreateEmptyFile(LastFile)
      else begin
        slFile := TStringList.Create;
        try
          slFile.Add(FmtDate(WF.DateEdit.Date) + '," "," ","' +
                     WF.DescEdit.Text + '"," "," ",' +
                     FmtMoney(Bal) + ',' + FmtMoney(Bal));
          slFile.SaveToFile(LastFile);
        finally
          slFile.Free;
        end;
      end;
      LoadFile(LastFile);
    end;
  finally
    WF.Free;
  end;
end;

function TMainForm.FmtDate(const ADate: TDateTime): string;
begin
  Result := AnsiDate(ADate); // FormatDateTime('yyyy-mm-dd', ADate);
end;

function TMainForm.FmtMoney(const AMoney: Currency): string;
begin
  Result := FormatCurr('#,##0.00', AMoney);
end;

function TMainForm.GetAmount: Currency;
begin
  if not TryStrToCurr(StringReplace(AmountEdit.Text, ',', '', [rfReplaceAll]), Result) then
    Result := -1.00
end;

procedure TMainForm.LoadFile(const AFile: string);
var
  sFile: string;
  slFile, slRow: TStringList;
  i, r: Integer;
  bConvert: Boolean;
begin

  MainGrid.RowCount := 2;
  MainGrid.Rows[1].Clear;

  // Set Column Headers
  MainGrid.Cells[ciDat, 0] := 'Date';
  MainGrid.Cells[ciChk, 0] := 'Check #';
  MainGrid.Cells[ciCat, 0] := 'Cat';
  MainGrid.Cells[ciTxt, 0] := 'Description';
  MainGrid.Cells[ciRec, 0] := '*';
  MainGrid.Cells[ciPmt, 0] := 'Payment';
  MainGrid.Cells[ciDep, 0] := 'Deposit';
  MainGrid.Cells[ciBal, 0] := 'Balance';

  {
  MainGrid.ColWidths[ciDat] := 70;
  MainGrid.ColWidths[ciChk] := 60;
  MainGrid.ColWidths[ciCat] := 45;
  MainGrid.ColWidths[ciTxt] := 300;
  MainGrid.ColWidths[ciRec] := 18;
  MainGrid.ColWidths[ciPmt] := 60;
  MainGrid.ColWidths[ciDep] := 60;
  MainGrid.ColWidths[ciBal] := 60;
  }
  for i := 0 to MainGrid.ColCount - 1 do
    MainGrid.ColWidths[i] := GetColWidth(i);

  MainGrid.FixedRows := 1;

  if AFile = '' then
    sFile := LastFile
  else
    sFile := AFile;
  if not FileExists(sFile) then
    Exit;

  if sFile = LastFile then
    AutoMonthlyBackup(sFile);

  slFile := TStringList.Create;
  slRow := TStringList.Create;
  try
    slFile.LoadFromFile(sFile);
    // Test the file to make sure it has the correct format:
    bConvert := False;
    if slFile.Count > 0 then begin
      slRow.CommaText := slFile[0];
      if slRow.Count = 7 then begin
        if not MsgOKCancel(sFile + #13#10 +
            'This Checkbook file does not contain the Category column.' + #13#10 +
            'Convert it to the current format?', 'Format Upgrade', Self) then
          Exit;
        bConvert := True;
      end;
    end;
    r := 0;
    for i := 0 to slFile.Count - 1 do begin
      if bConvert then begin
        slRow.CommaText := slFile[i];
        slRow.Insert(ciCat, ' ');
        slFile[i] := slRow.CommaText;
      end;
      if i > 0 then
        MainGrid.RowCount := MainGrid.RowCount + 1;
      Inc(r);
      MainGrid.Rows[r].CommaText := slFile[i];
    end;
  finally
    slFile.Free;
    slRow.Free;
  end;
  MainGrid.Row := MainGrid.RowCount - 1;
  MainGrid.Col := ciBal;

  if sFile <> LastFile then
    FBackupDone := False;
  LastFile := sFile;
  StatusBar.SimpleText := sFile;
  FDirty := bConvert;
  FMRU.Update(LastFile);
  Application.Title := 'Checkbook ' + ExtractFileNameNoExt(LastFile); // Taskbar Title
end;

procedure TMainForm.MoveCurrentRowDown;
var
  r: Integer;
begin
  r := MainGrid.Row;
  if (r > 0) and (r < MainGrid.RowCount - 1) then begin
    TStringGridHack(MainGrid).MoveRow(r, r + 1);
    MainGrid.Row := r + 1;
	FDirty := True;
  end;
end;

procedure TMainForm.MoveCurrentRowUp;
var
  r: Integer;
begin
  r := MainGrid.Row;
  if r > 1 then begin
    TStringGridHack(MainGrid).MoveRow(r, r - 1);
    MainGrid.Row := r - 1;
    FDirty := True;
  end;
end;

procedure TMainForm.MRUClicked(Sender: TObject);
var
  sFile: string;
begin
  // Sender is the menuitem that was clicked.
  sFile := TMenuItem(Sender).Caption;
  if FileExists(sFile) then begin
    if FDirty then
      RecalcAndSave;
    LoadFile(sFile);
  end else begin
    if MsgYesNo('The recent file "' + sFile + '" does not exist.' + #13#10 +
                'Delete it from the list?', 'File not Found', Self) then
      FMRU.Delete(sFile);
  end;
end;

procedure TMainForm.RecalcAndSave;
var
  Bal: Currency;
  r: Integer;
  sCat, sError: string;
  slMulti: TStringList;
begin
  Bal := 0;
  sError := '';
  slMulti := TStringList.Create;
  try
    slMulti.Delimiter := MultiDelim;
    for r := 1 to MainGrid.RowCount - 1 do begin
      if Length(MainGrid.Cells[ciPmt, r]) >= 4 {0.00} then
        Bal := Bal - CellAmount(ciPmt, r)
      else
        Bal := Bal + CellAmount(ciDep, r);
      MainGrid.Cells[ciBal, r] := FmtMoney(Bal);
      // Check for multi-category errors:
      sCat := MainGrid.Cells[ciCat, r];
      if Pos(MultiDelim, sCat) > 0 then begin
        slMulti.DelimitedText := sCat;
        if (slMulti[slMulti.Count - 1] = '') or
           (Pos('=', slMulti[slMulti.Count - 1]) > 0) then
          sError := sError + #13#10 + MainGrid.Cells[ciDat, r] + ' ' +
                                      MainGrid.Cells[ciTxt, r] + ': ' + sCat;
      end;
    end;
    SaveFile;
  finally
    slMulti.Free;
  end;
  if sError <> '' then
    MsgError('The following entries are missing remainder categories:' + sError,
             'Multi-Category Errors', Self);
end;

procedure TMainForm.ReportCategoryDetail(const AGroupByMonth, ASortByName: Boolean);
var
  slMonths, slCats, slEntries, slText, slMulti: TStringList;
  r, i, j: Integer;
  Amt, TotAmt: Currency;
  sDate, sCat, sAmt: string;
  procedure AddCatEntry;
  var
    CatTot: Currency;
    x: Integer;
  begin
    CatTot := Amt + StrToCurrDef(slCats.Values[sCat], 0.00);
    slCats.Values[sCat] := CurrToStr(CatTot);
    x := slCats.IndexOfName(sCat);
    if Assigned(slCats.Objects[x]) then
      slEntries := TStringList(slCats.Objects[x])
    else begin
      slEntries := TStringList.Create;
      slCats.Objects[x] := slEntries;
    end;
    // If this changes, SortEntriesByName must be updated to match:
    slEntries.Add(PadL(sAmt, 10) + '  ' +
                  MainGrid.Cells[ciDat, r] + '  ' +
                  MainGrid.Cells[ciTxt, r]);
  end;
  procedure SortEntriesByName;
  var
    x, y: Integer;
    bSorted: Boolean;
  begin
    for x := 0 to slEntries.Count - 1 do begin
      bSorted := True;
      for y := 0 to slEntries.Count - (x + 2) do begin
        if Copy(slEntries[y], 25, MaxInt) > Copy(slEntries[y + 1], 25, MaxInt) then begin
          slEntries.Exchange(y, y + 1);
          bSorted := False;
        end;
      end;
      if bSorted then
        Break;
    end;
  end;
begin
  slMonths := TStringList.Create;
  slText := TStringList.Create;
  slMulti := TStringList.Create;
  try
    slMulti.Delimiter := MultiDelim;
    // Gather totals:
    for r := 1 to MainGrid.RowCount - 1 do begin
      if AGroupByMonth then
        sDate := Copy(MainGrid.Cells[ciDat, r], 1, 7) // YYYY-MM
      else
        sDate := '';
      i := slMonths.IndexOf(sDate);
      if i < 0 then begin
        slCats := TStringList.Create;
        slCats.CaseSensitive := False; // Food == FOOD == food (last one found is what shows)
        slCats.Values[DepCat] := '0.00';
        slEntries := TStringList.Create;
        slMonths.AddObject(sDate, slEntries);
      end else
        slCats := TStringList(slMonths.Objects[i]);
      sCat := Trim(MainGrid.Cells[ciCat, r]);
      if Length(MainGrid.Cells[ciPmt, r]) >= 4 {0.00} then begin
        if sCat = '' then
          sCat := UnkCat;
        Amt := CellAmount(ciPmt, r);
        sAmt := MainGrid.Cells[ciPmt, r];
      end else if CategoryRefunds and (sCat <> '') then begin
        Amt := -CellAmount(ciDep, r);
        sAmt := '-' + MainGrid.Cells[ciDep, r];
      end else begin
        sCat := DepCat;
        Amt := CellAmount(ciDep, r);
        sAmt := MainGrid.Cells[ciDep, r];
      end;
      // Multi-category?
      if Pos(MultiDelim, sCat) > 0 then begin
        TotAmt := Amt;
        slMulti.DelimitedText := sCat;
        for i := 0 to slMulti.Count - 2 do begin // Skip the last one
          sCat := slMulti.Names[i];
          Amt := StrToCurrDef(slMulti.ValueFromIndex[i], 0.01);
          if TotAmt < 0 then
            Amt := -Amt;
          sAmt := FmtMoney(Amt);
          AddCatEntry;
          TotAmt := TotAmt - Amt;
        end;
        sCat := slMulti[slMulti.Count - 1];
        // Did the user forget the catch-all category?
        if (Pos('=', sCat) > 0) or (sCat = '') then begin
          MsgAlert('Missing remainder category on the following entry:' + #13#10 +
                   MainGrid.Cells[ciDat, r] + ' ' +
                   MainGrid.Cells[ciTxt, r] + ' ' +
                   MainGrid.Cells[ciCat, r],
                   'Multi-Category Error', Self);
          Exit;
        end;
        Amt := TotAmt; // Last category always gets the remainder
        sAmt := FmtMoney(Amt);
      end;
      AddCatEntry;
    end;
    // Write the output:
    slMonths.Sort;
    for r := 0 to slMonths.Count - 1 do begin
      slCats := TStringList(slMonths.Objects[r]);
      slCats.Sort;
      for i := 0 to slCats.Count - 1 do begin
        if AGroupByMonth then
          sAmt := slMonths[r] + ' ' + slCats.Names[i]
        else
          sAmt := slCats.Names[i];
        sAmt := sAmt + ' = ' + FmtMoney(StrToCurr(slCats.ValueFromIndex[i]));
        slText.Add(sAmt);
        slEntries := TStringList(slCats.Objects[i]);
        if ASortByName then
          SortEntriesByName;
        for j := 0 to slEntries.Count - 1 do
          slText.Add(slEntries[j]);
        slText.Add('');
      end;
    end;
    if AGroupByMonth then
      sCat := StripExt(LastFile) + ' Category Detail by Month.txt'
    else
      sCat := StripExt(LastFile) + ' Category Detail.txt';
    slText.SaveToFile(sCat);
    ShellExecute(Self.Handle, 'open', PChar(sCat), nil, nil, SW_SHOWNORMAL);
  finally
    for i := 0 to slMonths.Count - 1 do begin
      slCats := TStringList(slMonths.Objects[i]);
      for j := 0 to slCats.Count - 1 do
        TStringList(slCats.Objects[j]).Free; // slEntries
      slCats.Free;
    end;
    slMonths.Free;
    slText.Free;
    slMulti.Free;
  end;
end;

procedure TMainForm.ReportCategorySummary;
var
  slMonths, slCats, slTotals, slText, slMulti: TStringList;
  r, i, j, iLongCat, iAvgAdjust: Integer;
  Amt, Net, Avg, AvgNet: Currency;
  sDate, sCat, sVal: string;
  bDep: Boolean;
  procedure WriteCat(const AName: string; const AAvg: Currency = -12345678.90);
  begin
    if AAvg = -12345678.90 then
      slText.Add('  ' + PadR(AName, iLongCat) +
                 PadL(FmtMoney(Amt), 10))
    else
      slText.Add('  ' + PadR(AName, iLongCat) +
                 PadL(FmtMoney(Amt), 10) +
                 PadL(FmtMoney(AAvg), 12));
  end;
begin
  slMonths := TStringList.Create;
  slTotals := TStringList.Create;
  slText := TStringList.Create;
  slMulti := TStringList.Create;
  try
    slMulti.Delimiter := MultiDelim;
    // Gather totals:
    iLongCat := Length(NetCat);
    for r := 1 to MainGrid.RowCount - 1 do begin
      sDate := Copy(MainGrid.Cells[ciDat, r], 1, 7); // YYYY-MM
      i := slMonths.IndexOf(sDate);
      if i < 0 then begin
        slCats := TStringList.Create;
        slCats.CaseSensitive := False; // Food == FOOD == food (last one found is what shows)
        slCats.Values[DepCat] := '0.00';
        slMonths.AddObject(sDate, slCats);
      end else
        slCats := TStringList(slMonths.Objects[i]);
      sCat := Trim(MainGrid.Cells[ciCat, r]);
      (*
      if Length(MainGrid.Cells[ciPmt, r]) >= 4 {0.00} then begin
        if sCat = '' then
          sCat := UnkCat;
        Amt := StrToCurrDef(slCats.Values[sCat], 0.00) + CellAmount(ciPmt, r);
        slCats.Values[sCat] := CurrToStr(Amt);
      end else if CategoryRefunds and (sCat <> '') then begin
        // Treat deposits with a category as a credit/refund in that category
        Amt := StrToCurrDef(slCats.Values[sCat], 0.00) - CellAmount(ciDep, r);
        slCats.Values[sCat] := CurrToStr(Amt);
      end else begin
        sCat := DepCat;
        Amt := StrToCurrDef(slCats.Values[sCat], 0.00) + CellAmount(ciDep, r);
        slCats.Values[sCat] := CurrToStr(Amt);
      end;
      i := Length(sCat);
      if i > iLongCat then
        iLongCat := i;
      *)
      // Treat all as multi-category
      if Length(MainGrid.Cells[ciPmt, r]) >= 4 {0.00} then begin
        if sCat = '' then
          sCat := UnkCat;
        Amt := CellAmount(ciPmt, r);
      end else if CategoryRefunds and (sCat <> '') then begin
        // Treat deposits with a category as a credit/refund in that category
        Amt := -CellAmount(ciDep, r);
      end else begin
        sCat := DepCat;
        Amt := CellAmount(ciDep, r);
      end;
      slMulti.DelimitedText := sCat;
      for i := 0 to slMulti.Count - 1 do begin
        sCat := slMulti.Names[i];
        sVal := slMulti.ValueFromIndex[i];
        if sCat = '' then // The last (or only) category has no equal sign
          sCat := slMulti[i];
        j := Length(sCat);
        if j > iLongCat then
          iLongCat := j;
        if i < slMulti.Count - 1 then begin
          Net := StrToCurrDef(sVal, 0.01);
          if Amt < 0 then
            Net := -Net;
          Amt := Amt - Net;
        end else
          Net := Amt;
        slCats.Values[sCat] :=
          CurrToStr(StrToCurrDef(slCats.Values[sCat], 0.00) + Net);
      end;
    end;
    // Write the output:
    Inc(iLongCat); // Always one extra space
    slMonths.Sort;
    for r := 0 to slMonths.Count - 1 do begin
      slText.Add(slMonths[r]);
      slCats := TStringList(slMonths.Objects[r]);
      slCats.Sort;
      Net := 0;
      for i := 0 to slCats.Count - 1 do begin
        Amt := StrToCurr(slCats.ValueFromIndex[i]);
        if slCats.Names[i] = DepCat then
          Net := Net + Amt
        else
          Net := Net - Amt;
        WriteCat(slCats.Names[i]);
        slTotals.Values[slCats.Names[i]] :=
          CurrToStr(StrToCurrDef(slTotals.Values[slCats.Names[i]], 0.00) +
                    StrToCurrDef(slCats.ValueFromIndex[i], 0.00));
      end;
      Amt := Net;
      WriteCat(NetCat);
      slText.Add('');
    end;
    // If the first month is deposits only, ignore that month for calculating
    // average payment amounts.
    slCats := TStringList(slMonths.Objects[0]);
    if (slCats.Count = 1) and (slCats.Names[0] = DepCat) then
      iAvgAdjust := -1
    else
      iAvgAdjust := 0;
    slTotals.Sort;
    slText.Add(PadR('Totals:', iLongCat + 16) + 'Average:');
    Net := 0;
    AvgNet := 0;
    for i := 0 to slTotals.Count - 1 do begin
      bDep := (slTotals.Names[i] = DepCat);
      Amt := StrToCurr(slTotals.ValueFromIndex[i]);
      Avg := Amt / (slMonths.Count + Iif(bDep, 0, iAvgAdjust));
      WriteCat(slTotals.Names[i], Avg);
      if bDep then begin
        Net := Net + Amt;
        AvgNet := AvgNet + Avg;
      end else begin
        Net := Net - Amt;
        AvgNet := AvgNet - Avg;
      end;
    end;
    Amt := Net;
    WriteCat(NetCat, AvgNet);
    sCat := StripExt(LastFile) + ' Category Summary.txt';
    slText.SaveToFile(sCat);
    ShellExecute(Self.Handle, 'open', PChar(sCat), nil, nil, SW_SHOWNORMAL);
  finally
    for i := 0 to slMonths.Count - 1 do
      TStringList(slMonths.Objects[i]).Free;
    slMonths.Free;
    slTotals.Free;
    slText.Free;
    slMulti.Free;
  end;
end;

procedure TMainForm.ReportText;
var
  sFile, sStart, sEnd, sDate, sIndent: string;
  iChars, i, j: Integer;
  slText, slCats: TStringList;
  ColChars: array [ciDat..ciBal] of Integer;
  bStack: Boolean;
  function ColMax(const ACol: Integer): Integer;
  var
    iRow, iLen, iCat: Integer;
  begin
    Result := 0;
    if (ACol = ciCat) and bStack then begin
      for iRow := 0 to MainGrid.RowCount - 1 do begin
        slCats.DelimitedText := MainGrid.Cells[ACol, iRow];
        for iCat := 0 to slCats.Count - 1 do begin
          iLen := Length(slCats[iCat]);
          if iLen > Result then
            Result := iLen;
        end;
      end;
    end else begin
      for iRow := 0 to MainGrid.RowCount - 1 do begin
        iLen := Length(MainGrid.Cells[ACol, iRow]);
        if iLen > Result then
          Result := iLen;
      end;
    end;
  end;
begin
  {
  ciDat = 0; // Column Index constants
  ciChk = 1;
  ciCat = 2;
  ciTxt = 3;
  ciRec = 4;
  ciPmt = 5;
  ciDep = 6;
  ciBal = 7;
  }
  sFile := ChangeFileExt(LastFile, '.txt');
  if FileExists(sFile) then
    if not MsgOKCancel(sFile + ' exists.' + #13#10 + 'Overwrite?',
                       'File Exists', Self) then
      Exit;
  slText := TStringList.Create;
  slCats := TStringList.Create;
  try
    slCats.Delimiter := MultiDelim;
    iChars := TextWidth;
    bStack := TextStack;
    ColChars[ciDat] := 11; // yyyy-mm-dd
    if iChars > 0 then begin
      // Calculate widths relative to the visible grid:
      for i := ciChk to ciBal do
        ColChars[i] := Round((MainGrid.ColWidths[i] / MainGrid.Width) * iChars);
      ColChars[ciRec] := 1;  // Always * or blank
      for i := ciPmt to ciBal do begin // Amounts must always fit
        j := ColMax(i) + 1;
        if ColChars[i] < j then
          ColChars[i] := j;
      end;
      // Text absorbs any difference:
      j := 0;
      for i := ciDat to ciBal do
        j := j + ColChars[i];
      ColChars[ciTxt] := ColChars[ciTxt] - (j - iChars);
    end else begin
      // Make each column fit the longest data
      for i := ciChk to ciBal do
        ColChars[i] := ColMax(i) + 1;
      ColChars[ciRec] := 1;  // Always * or blank
    end;
    // Generate the text:
    slText.Add(LastFile);
    sStart := FSettings.Values['TextDate1'];
    sEnd := FSettings.Values['TextDate2'];
    if (sStart <> BlankSetting) or (sEnd <> BlankSetting) then
      slText.Add('Date Range: ' + sStart + ' to ' + sEnd);
    slText.Add('');
    if sStart = BlankSetting then
      sStart := '';
    if sEnd = BlankSetting then
      sEnd := '9999-99-99';
    sIndent := Spaces(ColChars[ciDat] + ColChars[ciChk]);
    for i := 0 to MainGrid.RowCount - 1 do begin
      sDate := MainGrid.Cells[ciDat, i];
      if (i = 0) or (sDate >= sStart) and (sDate <= sEnd) then begin
        if bStack then begin
          slCats.DelimitedText := MainGrid.Cells[ciCat, i];
          if slCats.Count = 0 then
            slCats.Add('');
          slText.Add(
            PadR(Ellipsis(MainGrid.Cells[ciDat, i], ColChars[ciDat]), ColChars[ciDat]) +
            PadR(Ellipsis(MainGrid.Cells[ciChk, i], ColChars[ciChk]), ColChars[ciChk]) +
            PadR(Ellipsis(slCats[0]               , ColChars[ciCat]), ColChars[ciCat]) +
            PadR(Ellipsis(MainGrid.Cells[ciTxt, i], ColChars[ciTxt]), ColChars[ciTxt]) +
            PadL(Ellipsis(MainGrid.Cells[ciRec, i], ColChars[ciRec]), ColChars[ciRec]) +
            PadL(Ellipsis(MainGrid.Cells[ciPmt, i], ColChars[ciPmt], False), ColChars[ciPmt]) +
            PadL(Ellipsis(MainGrid.Cells[ciDep, i], ColChars[ciDep], False), ColChars[ciDep]) +
            PadL(Ellipsis(MainGrid.Cells[ciBal, i], ColChars[ciBal], False), ColChars[ciBal]) );
          for j := 1 to slCats.Count - 1 do
            slText.Add(sIndent + slCats[j]);
        end else
          slText.Add(
            PadR(Ellipsis(MainGrid.Cells[ciDat, i], ColChars[ciDat]), ColChars[ciDat]) +
            PadR(Ellipsis(MainGrid.Cells[ciChk, i], ColChars[ciChk]), ColChars[ciChk]) +
            PadR(Ellipsis(MainGrid.Cells[ciCat, i], ColChars[ciCat]), ColChars[ciCat]) +
            PadR(Ellipsis(MainGrid.Cells[ciTxt, i], ColChars[ciTxt]), ColChars[ciTxt]) +
            PadL(Ellipsis(MainGrid.Cells[ciRec, i], ColChars[ciRec]), ColChars[ciRec]) +
            PadL(Ellipsis(MainGrid.Cells[ciPmt, i], ColChars[ciPmt], False), ColChars[ciPmt]) +
            PadL(Ellipsis(MainGrid.Cells[ciDep, i], ColChars[ciDep], False), ColChars[ciDep]) +
            PadL(Ellipsis(MainGrid.Cells[ciBal, i], ColChars[ciBal], False), ColChars[ciBal]) );
      end;
    end;
    // Save it and show it:
    slText.SaveToFile(sFile);
    ShellExecute(Self.Handle, 'open', PChar(sFile), nil, nil, SW_SHOWNORMAL);
  finally
    slText.Free;
    slCats.Free;
  end;
end;

procedure TMainForm.ResetFields(const AResetDate: Boolean);
begin
  DescEdit.Text := '';
  AmountEdit.Text := '0.00';
  cbDeposit.Checked := False;
  CheckNoEdit.Text := '';
  CatEdit.Text := DefaultCategory;
  if AResetDate then
    DateEdit.DateTime := Trunc(Now);
end;

procedure TMainForm.RollingBackup(const Force: Boolean);
var
  iMax, i: Integer;
  sPath, sName, sExt, sFrom, sTo: string;
begin
  if (Force or not FBackupDone) and (BackupCount > 0) then begin
    iMax := BackupCount;
    sPath := BackupDir;
    if Length(sPath) < 2 then
      sPath := ExtractFilePath(LastFile);
    sPath := AddSlash(sPath);
    sName := ExtractFileNameNoExt(LastFile);
    sExt := ExtractFileExt(LastFile);
    // Delete the last one:
    DeleteFile(sPath + sName + '.' + IntToStr(iMax - 1) + sExt);
    // Rename the rest down:
    for i := iMax - 2 downto 0 do begin
      sFrom := sPath + sName + '.' + IntToStr(i) + sExt;
      if FileExists(sFrom) then
        RenameFile(sFrom, sPath + sName + '.' + IntToStr(i + 1) + sExt);
    end;
    sFrom := LastFile;
    sTo := sPath + sName + '.0' + sExt;
    CopyFile(PChar(sFrom), PChar(sTo), False);
    FBackupDone := True; // Just once per editing session
  end;
end;

procedure TMainForm.SaveFile(const AFile: string);
var
  sFile: string;
  slFile: TStringList;
  r: Integer;
begin
  if AFile = '' then
    sFile := LastFile
  else
    sFile := AFile;
  if (sFile = LastFile) then
    RollingBackup;
  slFile := TStringList.Create;
  try
    for r := 1 to MainGrid.RowCount - 1 do
      slFile.Add(MainGrid.Rows[r].CommaText);
    slFile.SaveToFile(sFile);
    FDirty := False;
  finally
    slFile.Free;
  end;
end;

procedure TMainForm.SortDataByDate(const AForceRollingBackup: Boolean = False);
var
  slFile: TStringList;
  i, j: Integer;
  bSorted, bWasBackedUp: Boolean;
  c: TCursor;
begin
  c := Screen.Cursor;
  Screen.Cursor := crHourglass;
  try
	bWasBackedUp := FBackupDone;
    SaveFile;
    if AForceRollingBackup and (bWasBackedUp or not FBackupDone) then
      RollingBackup(AForceRollingBackup);
    slFile := TStringList.Create;
    try
      slFile.LoadFromFile(LastFile);
      { Unfortunately, CustomSort uses QuickSort, which randomly shuffles items
        that are "equal".  When I sort by date repeatedly, all the items for a
        particular day are in a different order each time.
        Good ol' bubblesort to the rescue!
      }
      for i := 0 to slFile.Count - 1 do begin
        bSorted := True;
        for j := 0 to slFile.Count - (i + 2) do begin
          if Copy(slFile[j], 1, 10) > Copy(slFile[j + 1], 1, 10) then begin
            slFile.Exchange(j, j + 1);
            bSorted := False;
          end;
        end;
        if bSorted then
          Break;
      end;
      slFile.SaveToFile(LastFile);
    finally
      slFile.Free;
    end;
    LoadFile();
    RecalcAndSave;
  finally
    Screen.Cursor := c;
  end;
end;

procedure TMainForm.SplitFileByDate(const ADate: TDateTime; const AName: string);
var
  sDateStr, sNewFile: string;
  slFile, slNew, slRow: TStringList;
  i: Integer;
  procedure SplitError(const AMsg: string);
  begin
    MsgAlert(AMsg, 'Split Cancelled', Self);
  end;
begin
  // Prep variables/sanity check:
  sDateStr := FmtDate(ADate);
  sNewFile := ExtractFilePath(LastFile) + AName + CheckFileExt;
  if SameText(sNewFile, LastFile) then begin
    SplitError('Current and New files are the same: ' + sNewFile);
    Exit;
  end;
  if FileExists(sNewFile) then begin
    SplitError('New file already exists: ' + sNewFile);
    Exit;
  end;
  // First save, backup and sort the current file:
  SortDataByDate(True);
  // Load the file into a stringlist and find the last line <= the cutoff date:
  slFile := TStringList.Create;
  slNew := TStringList.Create;
  slRow := TStringList.Create;
  try
    slFile.LoadFromFile(LastFile);
    slNew.Capacity := slFile.Count div 2;
    for i := slFile.Count - 1 downto 0 do begin
      if Copy(slFile[i], 1, 10) <= sDateStr then begin
        if slNew.Count = 0 then begin
          SplitError('No data would be moved to the new file.');
          Exit;
        end else if i = 0 then begin
          SplitError('Only one entry would remain in the current file.');
          Exit;
        end;
        // Get the final balance from the current file
        // and make it the first row of the new file:
        slRow.CommaText := slFile[i];
        slRow[ciDat] := sDateStr;
        slRow[ciChk] := ' ';
        slRow[ciCat] := ' ';
        slRow[ciTxt] := InitialEntry;
        slRow[ciPmt] := ' ';
        slRow[ciDep] := slRow[ciBal];
        slNew.Insert(0, slRow.CommaText);
        slFile.SaveToFile(LastFile);
        slNew.SaveToFile(sNewFile);
        LoadFile(sNewFile);
        break;
      end else begin
        // Move this line to the new file:
        slNew.Insert(0, slFile[i]);
        slFile.Delete(i);
      end;
    end;
    if slFile.Count = 0 then
      SplitError('All data would be moved to the new file.');
  finally
    slFile.Free;
    slNew.Free;
    slRow.Free;
  end;
end;

procedure TMainForm.StoreCurrentColWidths;
var
  i: Integer;
begin
  for i := 0 to MainGrid.ColCount - 1 do
    SetColWidth(i, MainGrid.ColWidths[i]);
end;

end.
