unit PrefsUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Spin, ComCtrls, StdCtrls, ShellAPI,
  llBool, llDates, llWinAPI, llMsgDlg;

type
  TPrefsForm = class(TForm)
    BackupCountSpin: TSpinEdit;
    BackupDirEdit: TEdit;
    Backups: TGroupBox;
    btnBackupDir: TButton;
    btnCancel: TButton;
    btnOK: TButton;
    CatEdit: TEdit;
    cbCategoryRefunds: TCheckBox;
    cbRequireCat: TCheckBox;
    dtpBackupDate: TDateTimePicker;
    InitialEntryEdit: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    btnResetCols: TButton;
    btnWebsite: TButton;
    procedure btnBackupDirClick(Sender: TObject);
    procedure btnOKClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure btnResetColsClick(Sender: TObject);
    procedure btnWebsiteClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    Settings: TStringList; // Provided by caller
  end;

implementation

{$R *.dfm}

uses llStrings, MainUnit; // For BlankSetting constant

procedure TPrefsForm.btnBackupDirClick(Sender: TObject);
var
  sDir: string;
begin
  sDir := BackupDirEdit.Text;
  if sDir = '' then
    sDir := ExtractFileDir(ParamStr(0));
  if BrowseForFolder(Self.Handle, 'Backup Directory', sDir) then
    BackupDirEdit.Text := sDir;
end;

procedure TPrefsForm.btnOKClick(Sender: TObject);
var
  sDir: string;
begin
  sDir := Trim(BackupDirEdit.Text);
  if (Length(sDir) > 1) and (dtpBackupDate.Checked or (BackupCountSpin.Value > 0)) then begin
    if not DirectoryExists(sDir) then begin
      MsgAlert('Backup directory does not exist.', 'Backup Directory', Self);
      BackupDirEdit.SetFocus;
      Exit;
    end;
  end;
  Settings.Values['AutoBackupDir'] := sDir;
  if dtpBackupDate.Checked then
    Settings.Values['AutoBackupDate'] := DtoS(dtpBackupDate.DateTime)
  else
    Settings.Values['AutoBackupDate'] := DtoS(0);
  Settings.Values['BackupCount'] := IntToStr(BackupCountSpin.Value);
  if IsBlank(CatEdit.Text) then
    Settings.Values['DefaultCategory'] := BlankSetting
  else
    Settings.Values['DefaultCategory'] := Trim(CatEdit.Text);
  Settings.Values['RequireCategory'] := BtoS(cbRequireCat.Checked);
  Settings.Values['InitialEntry'] := Trim(InitialEntryEdit.Text);
  Settings.Values['CategoryRefunds'] := BtoS(cbCategoryRefunds.Checked);
  ModalResult := mrOk;
end;

procedure TPrefsForm.btnResetColsClick(Sender: TObject);
begin
  MainForm.AssignDefaultColWidths;
end;

procedure TPrefsForm.btnWebsiteClick(Sender: TObject);
var
  sURL: string;
begin
  sURL := btnWebsite.Hint;
  ShellExecute(Application.MainForm.Handle, 'open', PChar(sURL),
               nil, nil, SW_SHOWNORMAL);
end;

procedure TPrefsForm.FormShow(Sender: TObject);
var
  dDate: TDateTime;
begin
  Assert(Assigned(Settings), 'TPrefsForm.Settings must be assigned before showing.');
  BackupDirEdit.Text := Settings.Values['AutoBackupDir'];
  dDate := StoDDef(Settings.Values['AutoBackupDate'], 0);
  if BackupDirEdit.Text = BlankSetting then begin
    BackupDirEdit.Text := '';
    dDate := 0;
  end;
  dtpBackupDate.DateTime := dDate;
  if dDate < 1 then begin
    dtpBackupDate.DateTime := FirstOfMonth(Date);
    dtpBackupDate.Checked := False;
  end;
  BackupCountSpin.Value := StrToIntDef(Settings.Values['BackupCount'], 0);
  CatEdit.Text := Settings.Values['DefaultCategory'];
  if CatEdit.Text = BlankSetting then
    CatEdit.Text := '';
  cbRequireCat.Checked := StoB(Settings.Values['RequireCategory']);
  cbCategoryRefunds.Checked := StoB(Settings.Values['CategoryRefunds'], True);
  InitialEntryEdit.Text := Settings.Values['InitialEntry'];
end;

end.
